@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -13,50 +13,53 @@
 ##############################################################################
 """Test suite for ZEO based on ZODB.tests."""
 
-# System imports
+from ZEO.ClientStorage import ClientStorage
+from ZEO.tests.forker import get_port
+from ZEO.tests import forker, Cache, CommitLockTests, ThreadTests
+from ZEO.tests import IterationTests
+from ZEO.zrpc.error import DisconnectedError
+from ZODB.tests import StorageTestBase, BasicStorage,  \
+     TransactionalUndoStorage,  \
+     PackableStorage, Synchronization, ConflictResolution, RevisionStorage, \
+     MTStorage, ReadOnlyStorage, IteratorStorage, RecoveryStorage
+from ZODB.tests.MinPO import MinPO
+from ZODB.tests.StorageTestBase import zodb_unpickle
+
 import asyncore
 import doctest
 import logging
 import os
+import persistent
+import shutil
 import signal
 import stat
+import sys
 import tempfile
 import threading
 import time
+import transaction
 import unittest
-import shutil
-
-# ZODB test support
+import ZEO.ServerStub
+import ZEO.StorageServer
+import ZEO.tests.ConnectionTests
+import ZEO.zrpc.connection
 import ZODB
 import ZODB.blob
+import ZODB.tests.testblob
 import ZODB.tests.util
-from ZODB.tests.MinPO import MinPO
-from ZODB.tests.StorageTestBase import zodb_unpickle
-import persistent
-import transaction
-
-# ZODB test mixin classes
-from ZODB.tests import StorageTestBase, BasicStorage, VersionStorage, \
-     TransactionalUndoStorage, TransactionalUndoVersionStorage, \
-     PackableStorage, Synchronization, ConflictResolution, RevisionStorage, \
-     MTStorage, ReadOnlyStorage
-
-from ZEO.ClientStorage import ClientStorage
-
-import ZEO.zrpc.connection
-
-from ZEO.tests import forker, Cache, CommitLockTests, ThreadTests
-from ZEO.tests.forker import get_port
-
-import ZEO.tests.ConnectionTests
-
-import ZEO.StorageServer
+import ZODB.utils
+import zope.testing.setupstack
 
 logger = logging.getLogger('ZEO.tests.testZEO')
 
+ZEO.zrpc.connection.start_client_thread()
+
 class DummyDB:
     def invalidate(self, *args):
         pass
+    def invalidateCache(*unused):
+        pass
+
 
 class CreativeGetState(persistent.Persistent):
     def __getstate__(self):
@@ -68,7 +71,7 @@ class MiscZEOTests:
     """ZEO tests that don't fit in elsewhere."""
 
     def checkCreativeGetState(self):
-        # This test covers persistent objects that provide their own 
+        # This test covers persistent objects that provide their own
         # __getstate__ which modifies the state of the object.
         # For details see bug #98275
 
@@ -116,6 +119,47 @@ class MiscZEOTests:
         finally:
             storage2.close()
 
+    def checkVolatileCacheWithImmediateLastTransaction(self):
+        # Earlier, a ClientStorage would not have the last transaction id
+        # available right after successful connection, this is required now.
+        addr = self._storage._addr
+        storage2 = ClientStorage(addr)
+        self.assert_(storage2.is_connected())
+        self.assertEquals(None, storage2.lastTransaction())
+        storage2.close()
+
+        self._dostore()
+        storage3 = ClientStorage(addr)
+        self.assert_(storage3.is_connected())
+        self.assertEquals(8, len(storage3.lastTransaction()))
+        self.assertNotEquals(ZODB.utils.z64, storage3.lastTransaction())
+        storage3.close()
+
+class ConfigurationTests(unittest.TestCase):
+
+    def checkDropCacheRatherVerifyConfiguration(self):
+        from ZODB.config import storageFromString
+        # the default is to do verification and not drop the cache
+        cs = storageFromString('''
+        <zeoclient>
+          server localhost:9090
+          wait false
+        </zeoclient>
+        ''')
+        self.assertEqual(cs._drop_cache_rather_verify, False)
+        cs.close()
+        # now for dropping
+        cs = storageFromString('''
+        <zeoclient>
+          server localhost:9090
+          wait false
+          drop-cache-rather-verify true
+        </zeoclient>
+        ''')
+        self.assertEqual(cs._drop_cache_rather_verify, True)
+        cs.close()
+
+
 class GenericTests(
     # Base class for all ZODB tests
     StorageTestBase.StorageTestBase,
@@ -129,7 +173,7 @@ class GenericTests(
     CommitLockTests.CommitLockVoteTests,
     ThreadTests.ThreadTests,
     # Locally defined (see above)
-    MiscZEOTests
+    MiscZEOTests,
     ):
 
     """Combine tests from various origins in one class."""
@@ -138,8 +182,9 @@ class GenericTests(
     blob_cache_dir = None
 
     def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
         logger.info("setUp() %s", self.id())
-        port = get_port()
+        port = get_port(self)
         zconf = forker.ZEOConfig(('', port))
         zport, adminaddr, pid, path = forker.start_zeo_server(self.getConfig(),
                                                               zconf, port)
@@ -148,7 +193,9 @@ class GenericTests(
         self._conf_path = path
         if not self.blob_cache_dir:
             # This is the blob cache for ClientStorage
-            self.blob_cache_dir = tempfile.mkdtemp()
+            self.blob_cache_dir = tempfile.mkdtemp(
+                'blob_cache',
+                dir=os.path.abspath(os.getcwd()))
         self._storage = ClientStorage(
             zport, '1', cache_size=20000000,
             min_disconnect_poll=0.5, wait=1,
@@ -158,14 +205,22 @@ class GenericTests(
 
     def tearDown(self):
         self._storage.close()
-        os.remove(self._conf_path)
-        ZODB.blob.remove_committed_dir(self.blob_cache_dir)
         for server in self._servers:
             forker.shutdown_zeo_server(server)
         if hasattr(os, 'waitpid'):
             # Not in Windows Python until 2.3
             for pid in self._pids:
                 os.waitpid(pid, 0)
+        StorageTestBase.StorageTestBase.tearDown(self)
+
+    def runTest(self):
+        try:
+            super(GenericTests, self).runTest()
+        except:
+            self._failed = True
+            raise
+        else:
+            self._failed = False
 
     def open(self, read_only=0):
         # Needed to support ReadOnlyStorage tests.  Ought to be a
@@ -188,15 +243,8 @@ class GenericTests(
         key = '%s:%s' % (self._storage._storage, self._storage._server_addr)
         self.assertEqual(self._storage.sortKey(), key)
 
-    def checkOmitVersionOnLoadAndHistory(self):
-        db = ZODB.DB(self._storage)
-        self.assertEqual(self._storage.load('\0'*8),
-                         self._storage.load('\0'*8, ''))
-        self._storage.history('\0'*8)
-
 class FullGenericTests(
     GenericTests,
-    Cache.StorageWithCache,
     Cache.TransUndoStorageWithCache,
     CommitLockTests.CommitLockUndoTests,
     ConflictResolution.ConflictResolvingStorage,
@@ -204,22 +252,93 @@ class FullGenericTests(
     PackableStorage.PackableUndoStorage,
     RevisionStorage.RevisionStorage,
     TransactionalUndoStorage.TransactionalUndoStorage,
-    TransactionalUndoVersionStorage.TransactionalUndoVersionStorage,
-    VersionStorage.VersionStorage,
+    IteratorStorage.IteratorStorage,
+    IterationTests.IterationTests,
     ):
     """Extend GenericTests with tests that MappingStorage can't pass."""
 
-class FileStorageTests(FullGenericTests):
-    """Test ZEO backed by a FileStorage."""
-    level = 2
+class FileStorageRecoveryTests(StorageTestBase.StorageTestBase,
+                               RecoveryStorage.RecoveryStorage):
 
     def getConfig(self):
-        filename = self.__fs_base = tempfile.mktemp()
         return """\
         <filestorage 1>
         path %s
         </filestorage>
-        """ % filename
+        """ % tempfile.mktemp(dir='.')
+
+    def _new_storage(self):
+        port = get_port(self)
+        zconf = forker.ZEOConfig(('', port))
+        zport, adminaddr, pid, path = forker.start_zeo_server(self.getConfig(),
+                                                              zconf, port)
+        self._pids.append(pid)
+        self._servers.append(adminaddr)
+
+        blob_cache_dir = tempfile.mkdtemp(dir='.')
+
+        storage = ClientStorage(
+            zport, '1', cache_size=20000000,
+            min_disconnect_poll=0.5, wait=1,
+            wait_timeout=60, blob_dir=blob_cache_dir)
+        storage.registerDB(DummyDB())
+        return storage
+
+    def setUp(self):
+        StorageTestBase.StorageTestBase.setUp(self)
+        self._pids = []
+        self._servers = []
+
+        self._storage = self._new_storage()
+        self._dst = self._new_storage()
+
+    def tearDown(self):
+        self._storage.close()
+        self._dst.close()
+
+        for server in self._servers:
+            forker.shutdown_zeo_server(server)
+        if hasattr(os, 'waitpid'):
+            # Not in Windows Python until 2.3
+            for pid in self._pids:
+                os.waitpid(pid, 0)
+        StorageTestBase.StorageTestBase.tearDown(self)
+
+    def new_dest(self):
+        return self._new_storage()
+
+
+class FileStorageTests(FullGenericTests):
+    """Test ZEO backed by a FileStorage."""
+
+    def getConfig(self):
+        return """\
+        <filestorage 1>
+        path Data.fs
+        </filestorage>
+        """
+
+    def checkInterfaceFromRemoteStorage(self):
+        # ClientStorage itself doesn't implement IStorageIteration, but the
+        # FileStorage on the other end does, and thus the ClientStorage
+        # instance that is connected to it reflects this.
+        self.failIf(ZODB.interfaces.IStorageIteration.implementedBy(
+            ZEO.ClientStorage.ClientStorage))
+        self.failUnless(ZODB.interfaces.IStorageIteration.providedBy(
+            self._storage))
+        # This is communicated using ClientStorage's _info object:
+        self.assertEquals(
+            (('ZODB.interfaces', 'IStorageRestoreable'),
+             ('ZODB.interfaces', 'IStorageIteration'),
+             ('ZODB.interfaces', 'IStorageUndoable'),
+             ('ZODB.interfaces', 'IStorageCurrentRecordIteration'),
+             ('ZODB.interfaces', 'IExternalGC'),
+             ('ZODB.interfaces', 'IStorage'),
+             ('zope.interface', 'Interface'),
+             ),
+            self._storage._info['interfaces']
+            )
+
 
 class MappingStorageTests(GenericTests):
     """ZEO backed by a Mapping storage."""
@@ -227,35 +346,37 @@ class MappingStorageTests(GenericTests):
     def getConfig(self):
         return """<mappingstorage 1/>"""
 
+    def checkSimpleIteration(self):
+        # The test base class IteratorStorage assumes that we keep undo data
+        # to construct our iterator, which we don't, so we disable this test.
+        pass
+
+    def checkUndoZombie(self):
+        # The test base class IteratorStorage assumes that we keep undo data
+        # to construct our iterator, which we don't, so we disable this test.
+        pass
+
 class DemoStorageTests(
     GenericTests,
-    Cache.StorageWithCache,
-    VersionStorage.VersionStorage,
     ):
 
     def getConfig(self):
         return """
         <demostorage 1>
           <filestorage 1>
-             path %s
+             path Data.fs
           </filestorage>
         </demostorage>
-        """ % tempfile.mktemp()
+        """
 
-    def checkLoadBeforeVersion(self):
-        # Doesn't implement loadBefore, except as a kind of place holder.
-        pass
-    
-    # the next three pack tests depend on undo
-
-    def checkPackVersionReachable(self):
+    def checkUndoZombie(self):
+        # The test base class IteratorStorage assumes that we keep undo data
+        # to construct our iterator, which we don't, so we disable this test.
         pass
 
-    def checkPackVersions(self):
-        pass
-
-    def checkPackVersionsInPast(self):
-        pass
+    def checkPackWithMultiDatabaseReferences(self):
+        pass # DemoStorage pack doesn't do gc
+    checkPackAllRevisions = checkPackWithMultiDatabaseReferences
 
 class HeartbeatTests(ZEO.tests.ConnectionTests.CommonSetupTearDown):
     """Make sure a heartbeat is being sent and that it does no harm
@@ -338,7 +459,7 @@ class CatastrophicClientLoopFailure(
             )
 
         ZEO.zrpc.connection.client_map[None] = Evil()
-        
+
         try:
             ZEO.zrpc.connection.client_trigger.pull_trigger()
         except DisconnectedError:
@@ -348,7 +469,7 @@ class CatastrophicClientLoopFailure(
         self.failIf(self._storage.is_connected())
         self.assertEqual(len(ZEO.zrpc.connection.client_map), 1)
         del ZEO.zrpc.connection.client_logger.critical
-        self.assertEqual(log[0][0], 'The ZEO cient loop failed.')
+        self.assertEqual(log[0][0], 'The ZEO client loop failed.')
         self.assert_('exc_info' in log[0][1])
         self.assertEqual(log[1][0], "Couldn't close a dispatcher.")
         self.assert_('exc_info' in log[1][1])
@@ -380,7 +501,7 @@ class ConnectionInvalidationOnReconnect(
                 self._invalidatedCache += 1
             def invalidate(*a, **k):
                 pass
-                
+
         db = DummyDB()
         storage.registerDB(db)
 
@@ -399,26 +520,22 @@ class ConnectionInvalidationOnReconnect(
 
         # Now, the root object in the connection should have been invalidated:
         self.assertEqual(db._invalidatedCache, base+1)
-    
+
 
 class CommonBlobTests:
 
-    def tearDown(self):
-        super(BlobAdaptedFileStorageTests, self).tearDown()
-        if os.path.exists(self.blobdir):
-            # Might be gone already if the super() method deleted
-            # the shared directory. Don't worry.
-            shutil.rmtree(self.blobdir)
-
     def getConfig(self):
         return """
         <blobstorage 1>
-          blob-dir %s
+          blob-dir blobs
           <filestorage 2>
-            path %s
+            path Data.fs
           </filestorage>
         </blobstorage>
-        """ % (self.blobdir, self.filestorage)
+        """
+
+    blobdir = 'blobs'
+    blob_cache_dir = 'blob_cache'
 
     def checkStoreBlob(self):
         from ZODB.utils import oid_repr, tid_repr
@@ -469,7 +586,6 @@ class CommonBlobTests:
              handle_serials
         import transaction
 
-        version = ''
         somedata = 'a' * 10
 
         blob = Blob()
@@ -502,24 +618,17 @@ class CommonBlobTests:
 
     def checkTransactionBufferCleanup(self):
         oid = self._storage.new_oid()
-        handle, blob_file_name = tempfile.mkstemp()
-        os.close(handle)
-        open(blob_file_name, 'w').write('I am a happy blob.')
+        open('blob_file', 'w').write('I am a happy blob.')
         t = transaction.Transaction()
         self._storage.tpc_begin(t)
         self._storage.storeBlob(
-          oid, ZODB.utils.z64, 'foo', blob_file_name, '', t)
+          oid, ZODB.utils.z64, 'foo', 'blob_file', '', t)
         self._storage.close()
 
 
-class BlobAdaptedFileStorageTests(GenericTests, CommonBlobTests):
+class BlobAdaptedFileStorageTests(FullGenericTests, CommonBlobTests):
     """ZEO backed by a BlobStorage-adapted FileStorage."""
 
-    def setUp(self):
-        self.blobdir = tempfile.mkdtemp()  # blob directory on ZEO server
-        self.filestorage = tempfile.mktemp()
-        super(BlobAdaptedFileStorageTests, self).setUp()
-
     def checkStoreAndLoadBlob(self):
         from ZODB.utils import oid_repr, tid_repr
         from ZODB.blob import Blob, BLOB_SUFFIX
@@ -532,7 +641,7 @@ class BlobAdaptedFileStorageTests(GenericTests, CommonBlobTests):
         for i in range(1000000):
             somedata.write("%s\n" % i)
         somedata.seek(0)
-        
+
         blob = Blob()
         bd_fh = blob.open('w')
         ZODB.utils.cp(somedata, bd_fh)
@@ -559,7 +668,7 @@ class BlobAdaptedFileStorageTests(GenericTests, CommonBlobTests):
         def check_data(path):
             self.assert_(os.path.exists(path))
             f = open(path, 'rb')
-            somedata.seek(0) 
+            somedata.seek(0)
             d1 = d2 = 1
             while d1 or d2:
                 d1 = f.read(8096)
@@ -571,7 +680,11 @@ class BlobAdaptedFileStorageTests(GenericTests, CommonBlobTests):
         check_data(filename)
 
         # ... and on the server
-        server_filename = filename.replace(self.blob_cache_dir, self.blobdir)
+        server_filename = os.path.join(
+            self.blobdir,
+            ZODB.blob.BushyLayout().getBlobFilePath(oid, revid),
+            )
+
         self.assert_(server_filename.startswith(self.blobdir))
         check_data(server_filename)
 
@@ -597,23 +710,14 @@ class BlobAdaptedFileStorageTests(GenericTests, CommonBlobTests):
             ]
         [thread.start() for thread in threads]
         [thread.join() for thread in threads]
-        [self.assertEqual(r, filename) for r in returns]        
+        [self.assertEqual(r, filename) for r in returns]
         check_data(filename)
-        
 
-class BlobWritableCacheTests(GenericTests, CommonBlobTests):
 
-    def setUp(self):
-        self.blobdir = self.blob_cache_dir = tempfile.mkdtemp()
-        self.filestorage = tempfile.mktemp()
-        self.shared_blob_dir = True
-        super(BlobWritableCacheTests, self).setUp()
-
-    def checkLoadNonExistingBlob(self):
-        # Regression test for #189542
-        self.assertRaises(ZODB.POSException.POSKeyError,
-                          self._storage.loadBlob, 'foo', 'foo')
+class BlobWritableCacheTests(FullGenericTests, CommonBlobTests):
 
+    blob_cache_dir = 'blobs'
+    shared_blob_dir = True
 
 class StorageServerClientWrapper:
 
@@ -647,9 +751,6 @@ class StorageServerWrapper:
     def supportsUndo(self):
         return False
 
-    def supportsVersions(self):
-        return False
-
     def new_oid(self):
         return self.server.new_oids(1)[0]
 
@@ -663,8 +764,8 @@ class StorageServerWrapper:
         del self.server.client.serials[:]
         return result
 
-    def store(self, oid, serial, data, version, transaction):
-        self.server.storea(oid, serial, data, version, id(transaction))
+    def store(self, oid, serial, data, version_ignored, transaction):
+        self.server.storea(oid, serial, data, id(transaction))
 
     def tpc_finish(self, transaction, func = lambda: None):
         self.server.tpc_finish(id(transaction))
@@ -675,7 +776,7 @@ def multiple_storages_invalidation_queue_is_not_insane():
     >>> from ZEO.StorageServer import StorageServer, ZEOStorage
     >>> from ZODB.FileStorage import FileStorage
     >>> from ZODB.DB import DB
-    >>> from persistent.dict import PersistentDict
+    >>> from persistent.mapping import PersistentMapping
     >>> from transaction import commit
     >>> fs1 = FileStorage('t1.fs')
     >>> fs2 = FileStorage('t2.fs')
@@ -690,24 +791,24 @@ def multiple_storages_invalidation_queue_is_not_insane():
     >>> commit()
     >>> o1 = conn1.root()
     >>> for i in range(10):
-    ...     o1.x = PersistentDict(); o1 = o1.x
+    ...     o1.x = PersistentMapping(); o1 = o1.x
     ...     commit()
 
     >>> last = fs1.lastTransaction()
     >>> for i in range(5):
-    ...     o1.x = PersistentDict(); o1 = o1.x
+    ...     o1.x = PersistentMapping(); o1 = o1.x
     ...     commit()
 
     >>> o2 = conn2.root()
     >>> for i in range(20):
-    ...     o2.x = PersistentDict(); o2 = o2.x
+    ...     o2.x = PersistentMapping(); o2 = o2.x
     ...     commit()
-    
+
     >>> trans, oids = s1.getInvalidations(last)
     >>> from ZODB.utils import u64
-    >>> sorted([int(u64(oid)) for (oid, v) in oids])
+    >>> sorted([int(u64(oid)) for oid in oids])
     [10, 11, 12, 13, 14]
-    
+
     >>> server.close_server()
     """
 
@@ -723,20 +824,20 @@ Let's create a file storage and stuff some data into it:
     >>> from ZEO.StorageServer import StorageServer, ZEOStorage
     >>> from ZODB.FileStorage import FileStorage
     >>> from ZODB.DB import DB
-    >>> from persistent.dict import PersistentDict
+    >>> from persistent.mapping import PersistentMapping
     >>> fs = FileStorage('t.fs')
     >>> db = DB(fs)
     >>> conn = db.open()
     >>> from transaction import commit
     >>> last = []
     >>> for i in range(100):
-    ...     conn.root()[i] = PersistentDict()
+    ...     conn.root()[i] = PersistentMapping()
     ...     commit()
     ...     last.append(fs.lastTransaction())
     >>> db.close()
 
 Now we'll open a storage server on the data, simulating a restart:
-    
+
     >>> fs = FileStorage('t.fs')
     >>> sv = StorageServer(('', get_port()), dict(fs=fs))
     >>> s = ZEOStorage(sv, sv.read_only)
@@ -759,7 +860,7 @@ structure using lastTransactions.
 
 
     >>> from ZODB.utils import u64
-    >>> sorted([int(u64(oid)) for (oid, version) in oids])
+    >>> sorted([int(u64(oid)) for oid in oids])
     [0, 92, 93, 94, 95, 96, 97, 98, 99, 100]
 
 (Note that the fact that we get oids for 92-100 is actually an
@@ -786,8 +887,8 @@ without this method:
     >>> sv = StorageServer(('', get_port()), dict(fs=fs))
     >>> st = StorageServerWrapper(sv, 'fs')
     >>> s = st.server
-    
-Now, if we ask fior the invalidations since the last committed
+
+Now, if we ask for the invalidations since the last committed
 transaction, we'll get a result:
 
     >>> tid, oids = s.getInvalidations(last[-1])
@@ -799,7 +900,7 @@ transaction, we'll get a result:
     >>> db = DB(st); conn = db.open()
     >>> ob = conn.root()
     >>> for i in range(5):
-    ...     ob.x = PersistentDict(); ob = ob.x
+    ...     ob.x = PersistentMapping(); ob = ob.x
     ...     commit()
     ...     last.append(fs.lastTransaction())
 
@@ -807,17 +908,20 @@ transaction, we'll get a result:
     >>> ntid == last[-1]
     True
 
-    >>> sorted([int(u64(oid)) for (oid, version) in oids])
+    >>> sorted([int(u64(oid)) for oid in oids])
     [0, 101, 102, 103, 104]
 
+    >>> fs.close()
     """
 
 def tpc_finish_error():
-    """Server errors in tpc_finish weren't handled properly.
+    r"""Server errors in tpc_finish weren't handled properly.
 
-    >>> import ZEO.ClientStorage
+    >>> import ZEO.ClientStorage, ZEO.zrpc.connection
 
     >>> class Connection:
+    ...     peer_protocol_version = (
+    ...         ZEO.zrpc.connection.Connection.current_protocol)
     ...     def __init__(self, client):
     ...         self.client = client
     ...     def get_addr(self):
@@ -834,6 +938,8 @@ def tpc_finish_error():
     ...         self.client = client
     ...     def connect(self, sync=1):
     ...         self.client.notifyConnected(Connection(self.client))
+    ...     def close(self):
+    ...         pass
 
     >>> class StorageServer:
     ...     should_fail = True
@@ -844,6 +950,8 @@ def tpc_finish_error():
     ...         return {}
     ...     def endZeoVerify(self):
     ...         self.conn.client.endVerify()
+    ...     def lastTransaction(self):
+    ...         return '\0'*8
     ...     def tpc_begin(self, t, *args):
     ...         if self.t is not None:
     ...             raise TypeError('already trans')
@@ -862,6 +970,8 @@ def tpc_finish_error():
     ...             raise TypeError('bad trans')
     ...         self.t = None
     ...         print 'abort'
+    ...     def iterator_gc(*args):
+    ...         pass
 
     >>> class ClientStorage(ZEO.ClientStorage.ClientStorage):
     ...     ConnectionManagerClass = ConnectionManager
@@ -905,29 +1015,305 @@ def tpc_finish_error():
     ... else: print "Should have failed"
     finish
     connection closed
+
+    >>> cs.close()
+    connection closed
     """
 
-test_classes = [
-    HeartbeatTests, CatastrophicClientLoopFailure,
-    ConnectionInvalidationOnReconnect, FileStorageTests,
-    MappingStorageTests, DemoStorageTests,
+def client_has_newer_data_than_server():
+    """It is bad if a client has newer data than the server.
+
+    >>> db = ZODB.DB('Data.fs')
+    >>> db.close()
+    >>> shutil.copyfile('Data.fs', 'Data.save')
+    >>> addr, admin = start_server(keep=1)
+    >>> db = ZEO.DB(addr, name='client', max_disconnect_poll=.01)
+    >>> wait_connected(db.storage)
+    >>> conn = db.open()
+    >>> conn.root().x = 1
+    >>> transaction.commit()
+
+    OK, we've added some data to the storage and the client cache has
+    the new data. Now, we'll stop the server, put back the old data, and
+    see what happens. :)
+
+    >>> stop_server(admin)
+    >>> shutil.copyfile('Data.save', 'Data.fs')
+
+    >>> import zope.testing.loggingsupport
+    >>> handler = zope.testing.loggingsupport.InstalledHandler(
+    ...     'ZEO', level=logging.ERROR)
+    >>> formatter = logging.Formatter('%(name)s %(levelname)s %(message)s')
+
+    >>> _, admin = start_server(addr=addr)
+
+    >>> for i in range(1000):
+    ...     while len(handler.records) < 5:
+    ...           time.sleep(.01)
+
+    >>> db.close()
+    >>> for record in handler.records[:5]:
+    ...     print formatter.format(record)
+    ... # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
+    ZEO.ClientStorage CRITICAL client
+    Client has seen newer transactions than server!
+    ZEO.zrpc ERROR (...) CW: error in notifyConnected (('localhost', ...))
+    Traceback (most recent call last):
+    ...
+    ClientStorageError: client Client has seen newer transactions than server!
+    ZEO.ClientStorage CRITICAL client
+    Client has seen newer transactions than server!
+    ZEO.zrpc ERROR (...) CW: error in notifyConnected (('localhost', ...))
+    Traceback (most recent call last):
+    ...
+    ClientStorageError: client Client has seen newer transactions than server!
+    ...
+
+    Note that the errors repeat because the client keeps on trying to connect.
+
+    >>> handler.uninstall()
+    >>> stop_server(admin)
+
+    """
+
+def history_over_zeo():
+    """
+    >>> addr, _ = start_server()
+    >>> db = ZEO.DB(addr)
+    >>> wait_connected(db.storage)
+    >>> conn = db.open()
+    >>> conn.root().x = 0
+    >>> transaction.commit()
+    >>> len(db.history(conn.root()._p_oid, 99))
+    2
+
+    >>> db.close()
+    """
+
+def dont_log_poskeyerrors_on_server():
+    """
+    >>> addr, admin = start_server()
+    >>> cs = ClientStorage(addr)
+    >>> cs.load(ZODB.utils.p64(1))
+    Traceback (most recent call last):
+    ...
+    POSKeyError: 0x01
+
+    >>> cs.close()
+    >>> stop_server(admin)
+    >>> 'POSKeyError' in open('server-%s.log' % addr[1]).read()
+    False
+    """
+
+def delete_object_multiple_clients():
+    """If we delete on one client, the delete should be reflected on the other.
+
+    First, we'll create an object:
+
+    >>> addr, _ = start_server()
+    >>> db = ZEO.DB(addr)
+    >>> conn = db.open()
+    >>> conn.root()[0] = conn.root().__class__()
+    >>> transaction.commit()
+    >>> oid = conn.root()[0]._p_oid
+
+    We verify that we can read it in another client, which also loads
+    it into the client cache.
+
+    >>> cs = ClientStorage(addr)
+    >>> p, s = cs.load(oid)
+
+    Now, we'll remove the object:
+
+    >>> txn = transaction.begin()
+    >>> db.storage.tpc_begin(txn)
+    >>> db.storage.deleteObject(oid, s, txn)
+    >>> db.storage.tpc_vote(txn)
+    >>> db.storage.tpc_finish(txn)
+
+    And we'll get a POSKeyError if we try to access it:
+
+    >>> db.storage.load(oid) # doctest: +ELLIPSIS
+    Traceback (most recent call last):
+    ...
+    POSKeyError: ...
+
+    We'll wait for our other storage to get the invalidation and then
+    try to access the object. We'll get a POSKeyError there too:
+
+    >>> tid = db.storage.lastTransaction()
+    >>> forker.wait_until(
+    ...    'cs has caught up',
+    ...    lambda : cs.lastTransaction() == tid)
+    >>> cs.load(oid) # doctest: +ELLIPSIS
+    Traceback (most recent call last):
+    ...
+    POSKeyError: ...
+
+    >>> db.close()
+    >>> cs.close()
+    """
+
+
+def open_convenience():
+    """Often, we just want to open a single connection.
+
+    >>> addr, _ = start_server(path='data.fs')
+    >>> conn = ZEO.connection(addr)
+    >>> conn.root()
+    {}
+
+    >>> conn.root()['x'] = 1
+    >>> transaction.commit()
+    >>> conn.close()
+
+    Let's make sure the database was cloased when we closed the
+    connection, and that the data is there.
+
+    >>> db = ZEO.DB(addr)
+    >>> conn = db.open()
+    >>> conn.root()
+    {'x': 1}
+    >>> db.close()
+    """
+
+def client_asyncore_thread_has_name():
+    """
+    >>> len([t for t in threading.enumerate()
+    ...      if t.getName() == 'ZEO.zrpc.connection'])
+    1
+    """
+
+def runzeo_without_configfile():
+    """
+    >>> open('runzeo', 'w').write('''
+    ... import sys
+    ... sys.path[:] = %r
+    ... import ZEO.runzeo
+    ... ZEO.runzeo.main(sys.argv[1:])
+    ... ''' % sys.path)
+
+    >>> import subprocess, re
+    >>> print re.sub('\d\d+|[:]', '', subprocess.Popen(
+    ...     [sys.executable, 'runzeo', '-a:%s' % get_port(), '-ft', '--test'],
+    ...     stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
+    ...     ).stdout.read()), # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
+    ------
+    --T INFO ZEO.runzeo () opening storage '1' using FileStorage
+    ------
+    --T INFO ZEO.StorageServer () StorageServer created RW with storages 1RWt
+    ------
+    --T INFO ZEO.zrpc () listening on ...
+    ------
+    --T INFO ZEO.runzeo () closing storage '1'
+    testing exit immediately
+    """
+
+slow_test_classes = [
     BlobAdaptedFileStorageTests, BlobWritableCacheTests,
+    DemoStorageTests, FileStorageTests, MappingStorageTests,
     ]
 
+quick_test_classes = [
+    FileStorageRecoveryTests, ConfigurationTests, HeartbeatTests,
+    CatastrophicClientLoopFailure, ConnectionInvalidationOnReconnect,
+    ]
+
+class ServerManagingClientStorage(ClientStorage):
+
+    class StorageServerStubClass(ZEO.ServerStub.StorageServer):
+
+        # Wait for abort for the benefit of blob_transacton.txt
+        def tpc_abort(self, id):
+            self.rpc.call('tpc_abort', id)
+
+    def __init__(self, name, blob_dir, shared=False, extrafsoptions=''):
+        if shared:
+            server_blob_dir = blob_dir
+        else:
+            server_blob_dir = 'server-'+blob_dir
+        self.globs = {}
+        port = forker.get_port2(self)
+        addr, admin, pid, config = forker.start_zeo_server(
+            """
+            <blobstorage>
+                blob-dir %s
+                <filestorage>
+                   path %s
+                   %s
+                </filestorage>
+            </blobstorage>
+            """ % (server_blob_dir, name+'.fs', extrafsoptions),
+            port=port,
+            )
+        os.remove(config)
+        zope.testing.setupstack.register(self, os.waitpid, pid, 0)
+        zope.testing.setupstack.register(
+            self, forker.shutdown_zeo_server, admin)
+        if shared:
+            ClientStorage.__init__(self, addr, blob_dir=blob_dir,
+                                   shared_blob_dir=True)
+        else:
+            ClientStorage.__init__(self, addr, blob_dir=blob_dir)
+
+    def close(self):
+        ClientStorage.close(self)
+        zope.testing.setupstack.tearDown(self)
+
+def create_storage_shared(name, blob_dir):
+    return ServerManagingClientStorage(name, blob_dir, True)
+
+class ServerManagingClientStorageForIExternalGCTest(
+    ServerManagingClientStorage):
+
+    def pack(self, t=None, referencesf=None):
+        ServerManagingClientStorage.pack(self, t, referencesf, wait=True)
+        # Packing doesn't clear old versions out of zeo client caches,
+        # so we'll clear the caches.
+        self._cache.clear()
+        ZEO.ClientStorage._check_blob_cache_size(self.blob_dir, 0)
+
 def test_suite():
     suite = unittest.TestSuite()
-    suite.addTest(doctest.DocTestSuite(setUp=ZODB.tests.util.setUp,
-                                       tearDown=ZODB.tests.util.tearDown))
-    suite.addTest(doctest.DocFileSuite('registerDB.test'))
-    suite.addTest(
-        doctest.DocFileSuite('zeo-fan-out.test',
-                             setUp=ZODB.tests.util.setUp,
-                             tearDown=ZODB.tests.util.tearDown,
-                             ),
+
+    # Collect misc tests into their own layer to educe size of
+    # unit test layer
+    zeo = unittest.TestSuite()
+    zeo.addTest(unittest.makeSuite(ZODB.tests.util.AAAA_Test_Runner_Hack))
+    zeo.addTest(doctest.DocTestSuite(
+        setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown))
+    zeo.addTest(doctest.DocTestSuite(ZEO.tests.IterationTests,
+        setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown))
+    zeo.addTest(doctest.DocFileSuite('registerDB.test'))
+    zeo.addTest(
+        doctest.DocFileSuite(
+            'zeo-fan-out.test', 'zdoptions.test',
+            'drop_cache_rather_than_verify.txt', 'client-config.test',
+            'protocols.test', 'zeo_blob_cache.test', 'invalidation-age.txt',
+            setUp=forker.setUp, tearDown=zope.testing.setupstack.tearDown,
+            ),
         )
-    for klass in test_classes:
+    zeo.addTest(PackableStorage.IExternalGC_suite(
+        lambda :
+        ServerManagingClientStorageForIExternalGCTest(
+            'data.fs', 'blobs', extrafsoptions='pack-gc false')
+        ))
+    for klass in quick_test_classes:
+        zeo.addTest(unittest.makeSuite(klass, "check"))
+    zeo.layer = ZODB.tests.util.MininalTestLayer('testZeo-misc')
+    suite.addTest(zeo)
+
+    # Put the heavyweights in their own layers
+    for klass in slow_test_classes:
         sub = unittest.makeSuite(klass, "check")
+        sub.layer = ZODB.tests.util.MininalTestLayer(klass.__name__)
         suite.addTest(sub)
+
+    suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
+        'ClientStorageNonSharedBlobs', ServerManagingClientStorage))
+    suite.addTest(ZODB.tests.testblob.storage_reusable_suite(
+        'ClientStorageSharedBlobs', create_storage_shared))
+
     return suite
 
 
