@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
+# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -12,9 +12,6 @@
 #
 ##############################################################################
 
-# The next line must use double quotes, so release.py recognizes it.
-__version__ = "3.7.0b3"
-
 import sys
 
 from persistent import TimeStamp
@@ -28,4 +25,4 @@ sys.modules['ZODB.PersistentList'] = sys.modules['persistent.list']
 
 del mapping, list, sys
 
-from DB import DB
+from DB import DB, connection
